from capytale.autoeval import validationclass

# =============================================================================
# Une classe de validation pour vérifier la présence des attributs et méthodes
# =============================================================================

@validationclass
class ValidateClassContent:
    """Permet de tester le contenu d'un objet en fonction des pamateres passés a l'instanciation.
    """
    def __init__(self, class_name, parameters_init, attributs, methods):
        """
        parameters
        ----------
        class_name : str
            le nom de la classe a tester
            
        parameters_init : list de tuple
            Les parametres a passer lors des instanciations. 
            Chaque élément de la liste est un tuple contenant les parametres a transmettre.
            Si pas de paramètre, mettre un tuple vide.
            
        attributs : dict
            le nom de chaque attribut (str) est une clé du dictionnaire. Les valeur sont des listes contenant
            les valeurs successives prisent par l'attribut lors des instaciations successives.
            
        methods : list
            La liste des noms des méthodes a vérifier.
        """
        super().__init__()
        self.class_name = class_name
        self.attributs = attributs
        self.methods = methods
        self.parametres = parameters_init
        

    def __call__(self):
        flag_error = False
        for i in range(len(self.parametres)):
            try:
                import __main__
                # instancie l'objet
                _class = getattr(__main__, self.class_name)                
                instance = _class(*self.parametres[i])
                
                # test du contenu de la classe
                # les bons attributs et leur valeur
                for a in self.attributs:
                    if hasattr(instance, a):
                        print(f"✅ l'attribut {a} est bien déclaré")
                        att_a = getattr(instance, a)
                        if att_a is self.attributs[a][i]:
                            print(f"✅ l'attribut {a} reçoit la bonne valeur {att_a}")
                        else:
                            print(f"❌ l'attribut {a} ne reçoit pas la bonne valeur : {self.attributs[a][i]} et non {att_a}")
                            flag_error = True
                    else:
                        print(f"❌ l'attribut {a} n'est pas déclaré")
                        flag_error = True

                for m in self.methods:
                    if hasattr(instance, m):
                        print(f"✅ la méthode {m} est bien déclarée")
                    else:
                        print(f"❌ la méthode {m} n'est pas déclarée")            
                        flag_error = True    
                        
            except AssertionError:
                print(f"❌ AssertionError avec les parametres {self.parametres[i]}")
        return not flag_error

# =============================================================================
# Une classe de validation pour tester une méthode
# =============================================================================
@validationclass
class ValidateMethod:
    """Permet de tester une classe.
    """
    def __init__(self, class_name=None, parameters_init=None, instance=None, method_name=None, 
                 test_values=None, return_values=None, method_verif=None, return_values_verif=None):
        """
        parameters
        ----------
        class_name : str
            le nom de la classe a tester. Valeur par défaut None. Si fourni l'objet sera instancié à partir de cette classe.
            
        parameters_init : tuple
            Les parametres a passer lors de l'instanciation. 
            
        instance : str
            nom d'une instance de la classe a tester, par défaut None. Si défini, c'est cette instance qui sera testée.
            
        method_name : str
            le nom de la méthode a tester
            
        test_values : list de tuple
            Chaque tuple contient les parametres de la méthode (method_name) a passer lors du test. 
            Un tuple vide attendu si aucun parametre.
            
        return_values : list
            Chaque élément de la list est la valeur de retour attendu du test correspondant.
            
        method_verif : str
            Le nom d'une méthode ou d'un attribut a tester en plus. Si c'est une méthode, elle ne doit pas demander
            d'argument. Permet de tester l'état de l'objetaprès exécution de method_name. Optionnel, par defaut None.
            
        return_values_verif : list
            La liste des valeur de retour attendues lors de l'exécution de method_verif.
        """
        
        super().__init__()
        self.class_name = class_name
        self.parameters_init = parameters_init
        self.instance = instance
        self.method_name = method_name
        self.test_values = test_values
        self.return_values = return_values
        self.return_values_verif = return_values_verif
        self.method_verif_name = method_verif

            
    def __call__(self):
        # instancie un objet de la classe ou récupère directement une instance
        import __main__
        if self.instance is None:
            if self.class_name is not None and self.parameters_init is not None:
                _class = getattr(__main__, self.class_name)
                inst = _class(*self.parameters_init)                
            else:
                raise NameError('Il manque des arguments ...')
        else:
            inst = getattr(__main__, self.instance)    
            
        verif = False
        if self.method_verif_name is not None and self.return_values_verif is not None:
            self.method_verif = getattr(inst, self.method_verif_name)
            verif = True            
           
        # essaye de récupérer la méthode a tester
        try:
            self.method = getattr(inst, self.method_name)
        except:
            print(f"❌ {self.method_name} ne semble pas exister dans votre classe.")
            return False
        
        # début des tests
        flag_error = False
        for i in range(len(self.test_values)):
            # test des valeurs de retour de la méthode
            value_returned = self.method(*self.test_values[i])
            if value_returned == self.return_values[i]:
                print(f"✅ la méthode {self.method.__name__}{self.test_values[i]} renvoi bien {value_returned}.")
            else:
                print(f"❌ pour des parametres d'entrée {self.test_values[i]}, la méthode renvoi {value_returned} au lieu de {self.return_values[i]} !")
                flag_error = True 
            
            if verif:
                # vérification de l'état de lobjet par une autre méthode ou un attribut: method_verif
                if callable(self.method_verif):
                    value_verif = self.method_verif()
                else:
                    #il s'agit d'un attribut
                    value_verif = self.method_verif
                if value_verif == self.return_values_verif[i]:
                    print(f"✅ la méthode {self.method_verif.__name__} renvoi bien {value_verif}")
                else:
                    print(f"❌ la méthode de vérification renvoi {value_verif} au lieu de {self.return_values_verif[i]} !")
                    flag_error = True  
        return not flag_error