from contextlib import contextmanager
from typing import List, Tuple, Union
import inspect


class Tests:
    def __init__(self, total: int = 0):
        self.passed: int = 0
        self.failed: int = 0
        self.total: int = total
        self.messages: List[Tuple[bool, str]] = []

    @property
    def num(self):
        return self.passed + self.failed

    @contextmanager
    def exception(
        self,
        exception: Exception,
        message_ok: Union[str, None] = None,
        message_ko: Union[str, None] = None,
        fail_on_error: bool = False,
    ):
        try:
            yield
            self.test(False, message_ok, message_ko, fail_on_error)
        except exception:
            self.test(True, message_ok, message_ko, fail_on_error)

    def test(
        self,
        result: bool,
        message_ok: Union[str, None] = None,
        message_ko: Union[str, None] = None,
        stop_on_fail: bool = False,
    ):
        if result:
            self.passed += 1
            self.messages.append(
                (
                    True,
                    (
                        message_ok
                        if message_ok is not None
                        else f"Test {self.num} passé avec succès"
                    ),
                )
            )
        else:
            self.failed += 1
            self.messages.append(
                (
                    False,
                    (
                        message_ko
                        if message_ko is not None
                        else (
                            message_ok
                            if message_ok is not None
                            else f"Test {self.num} : échec du test"
                        )
                    ),
                )
            )
            if stop_on_fail:
                raise AssertionError("NORMAL")

    def fail(
        self,
        result: bool,
        message_ok: Union[str, None] = None,
        message_ko: Union[str, None] = None,
    ):
        self.test(result, message_ok, message_ko, True)

    def end(self, exception=None):
        def print_err(skk):
            print("\033[91m {}\033[00m".format(skk))

        def print_success(skk):
            print("\033[92m {}\033[00m".format(skk))

        if exception is not None:
            if exception.args[0] != "NORMAL":
                print(f"❌ Erreur inconnue : {exception.args[0]}")
        if self.total == 0:
            self.total = self.passed + self.failed
        print(f"🛈 {self.total} TEST(S) EXÉCUTÉ(S)", end=" : ")
        if any([not success for success, _ in self.messages]):
            print_err(f"❌ {self.failed}/{self.total} tests ont échoué")
        else:
            print_success(f"✅ Tous les tests ont été passés avec succès")
        for success, message in self.messages:
            if success:
                print_success(f"✅ : {message}")
            else:
                print_err(f"❌ : {message}")
        print()


def test_exercice():
    t = Tests()
    g = inspect.stack()[1].frame.f_globals
    try:
        t.fail(
            "somme" in g,
            "La fonction `somme` existe",
            "La fonction `somme` n'existe pas",
        )
        somme = g["somme"]
        sig = inspect.signature(somme)
        t.fail(
            len(sig.parameters) == 1,
            "La fonction `somme` prend un argument",
        )
        t.fail(
            "tab" in sig.parameters,
            "La fonction `somme` prend un argument `tab`",
        )
        t.test(
            sig.parameters["tab"].default == inspect._empty,
            "L'argument `tab` n'a pas de valeur par défaut",
        )

        # Si on veut vérifier le typage des arguments (pas au programme)
        # t.fail(
        #     sig.parameters["tab"].annotation == List[int],
        #     "Le premier argument de `somme` est de type `List[int]`",
        #     "Le premier argument de `somme` n'est pas de type `List[int]`",
        # )

        # Test des préconditions
        with t.exception(
            AssertionError,
            "Vérification que `tab` est une liste",
        ):
            somme(0)
        with t.exception(
            AssertionError,
            "Vérification que `tab` contient uniquement des entiers",
        ):
            somme("test")
        with t.exception(
            AssertionError,
            "Vérification que `tab` est non vide",
        ):
            somme([])

        # Vérification des résultats
        tests_valeurs = [
            ([0, 1, 2], 3),
            ([0, 1, 2, 250], 3),
            ([0, 1, 2, 100], 103),
            ([0, 1, 2, 100, 101], 103),
        ]
        for arg, result in tests_valeurs:
            t.test(
                somme(tab=arg) == result,
                f"La somme de {arg} est {result}",
                f"La somme de {arg} n'est pas {result}",
            )

    except AssertionError as e:
        t.end(e)
    else:
        t.end()
